#!/usr/bin/python

# TODO:
# - maybe extrude the Add/Remove/Scan stuff into an external file to include in both netdbg and connect?
# - maybe a Latency Test button or at least an estimate of a theoretical link latency

# Author: Leonardo Gabrielli (l.gabrielli@univpm.it)
# Created: 2014/10/28
# License: GPLv3

# treeModel:
# +
# |
# +--node1(column1, column2)
# |
# +--node2(column1, column2)
#
# a node is referenced using a path, a path can be converted to a treeiter by get_iter()



WEMUST_CONNECT_VERSION = "0.1.0"

DRAWING_WIDTH = 100

# ---------------- IMPORTS -------------------- #

from gi.repository import Gtk, Gdk
from gi.repository import GObject as gobject
import cairo
import threading
from wemust import USE_NETDBG, SERVICETYPE, UUID_LEN, NETDBG_MCAST_GROUP, NETDBG_BASE_PORT, NETDBG_PRINT_PORT, NETDBG_JACKD_PORT, \
    NETDBG_JTSER_PORT, NETDBG_JTCLI_PORT, NETDBG_CTRL_PORT, MAX_CTRL_CMD_LEN, CTRL_CMD_START, CTRL_CMD_STOP, CTRL_CMD_HALT, \
    CTRL_CMD_REBOOT, WEMUST_GUTS_VERSION, HOSTNAMETYPE, IPv4TYPE, IPv6TYPE
import socket
import struct
import time
import sys, os
import subprocess
import re
from collections import deque

# ---------------- PATHS -------------------- #

SRC_PATH=(os.path.split(os.path.realpath(__file__)))[0] + '/'
HELP_FILE_PATH = SRC_PATH + "docs/help.txt"

# ---------------- INITS -------------------- #

globalQLock = threading.Lock() # one lock for all the queues: reduced parallelism... to change
dbgQ = deque()
jtcliQ = deque()
jtserQ = deque()
jdQ = deque()

gobject.threads_init()

threadsList = []
toKill = None
quitTimeout = 10 #seconds
quitFlag = threading.Event()


# ------------ UTILITY FUNCS --------------- #

def isValidIP(addr):
    if ( addr ):
        try:
            socket.inet_pton(addr, socket.AF_INET) # AF_INET6 for future IPv6 compat
            return 1
        except socket.error:
            return 0
    else:
        return 0

def isValidHostname(addr):
    return 1 # to implement (some check on irregular chars?)

def isValidHTMLColor(colorStr):
    if ( re.match("^#(?:[0-9a-fA-F]{3}){1,2}$", colorStr) != None ):
        return 1
    else:
        return 0

# ---------------------------------------------------------- #
# ---------------- THREAD HANDLER CLASS -------------------- #
# ---------------------------------------------------------- #

class threadHandler(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)

    def run(self):
        try:
            while( not quitFlag.is_set() ):
                if( toKill == None ):
                    time.sleep( 2 )
                else:
                    toKill.join(quitTimeout)

        except (KeyboardInterrupt, SystemExit):
            print("Keyboard Interrupt caught: exiting")
            raise


# ---------------------------------------------------------- #
# ----------------   NODE ENTRY CLASS   -------------------- #
# ---------------------------------------------------------- #

class nodeEntry:

    def __init__(self, hostname, IP):
        self.hostname = hostname
        self.IP = IP

    # connect to the IP in treeModel[iterator]
    def addConnection(self, treeModel, iterator): # get the iterator in a tree of nodeEntry items
        path = treeModel.get_string_from_iter(iterator)
        peerIP = treeModel.get_value(iterator, 1)
        # check if already connected

        # call remotely jacktrip -c peerIP and jack_connect
        # now process the path in order to find the row number, needed for drawing the line
        # draw

    def delConnection(self, treeModel, iterator):
        path = treeModel.get_string_from_iter(iterator)
        peerIP = treeModel.get_value(iterator, 1)
        # kill remotely jacktrip -c peerIP
        # now process the path in order to find the row number, needed for drawing the line
        # delete line

# ---------------------------------------------------------- #
# ---------------- GUI CLASS FOR NETDBG -------------------- #
# ---------------------------------------------------------- #

class connectGUI:
    wTree = Gtk.Builder()
    clearSema = threading.Semaphore(0)

    def __init__(self):
        self.builder = Gtk.Builder()
        self.builder.add_from_file(SRC_PATH + "GUI/wemust-connect.glade")
        self.builder.add_from_file(SRC_PATH + "GUI/add-dialog.glade")
        self.builder.add_from_file(SRC_PATH + "GUI/netdbg-about.glade")
        self.builder.add_from_file(SRC_PATH + "GUI/netdbg-help.glade")
        # GUI init
        self.getAll()
        self.connectAll()
        self.initTreeViews()
        file = open(SRC_PATH + 'wemust-netdbg.py', "r")
        we_net = '-'
        for line in file:
            if re.search('WEMUST_NETDBG_VERSION', line):
                we_net = re.search('\"(.+?)\"', line)
                we_net = we_net.group(1)
        self.aboutDialog.set_version("VERSIONING:\n wemust-connect: v." + WEMUST_CONNECT_VERSION + "\n wemust-netdbg: v." + we_net + "\n wemust.py lib: v." + WEMUST_GUTS_VERSION)


    # GET ALL GUI OBJECTS FROM THE GLADE FILE
    def getAll(self):
        self.window = self.builder.get_object("connectWindow")
        #
        self.addB = self.builder.get_object("addButton")
        self.remB = self.builder.get_object("removeButton")
        self.scanB = self.builder.get_object("scanButton")
        self.dbgB = self.builder.get_object("dbgButton")
        self.sshQJackB = self.builder.get_object("sshQJackButton")
        #
        self.deviceTree = self.builder.get_object("deviceTree")
        self.leftDeviceView = self.builder.get_object("leftDeviceView")
        self.rightDeviceView = self.builder.get_object("rightDeviceView")
        #
        self.linesArea = self.builder.get_object("linesArea")
        self.linesArea.set_size_request(100, 100)

        #self.testo  = self.builder.get_object("dbgPrintTerm")

        self.aboutItem = self.builder.get_object("connectAbout")
        self.helpItem = self.builder.get_object("connectHelp")

        # ADD DIALOG
        self.addDialog = self.builder.get_object("addDialog")
        self.addDialogButtonBox = self.builder.get_object("addDialogButtonBox")
        self.addDialogApplyButton = self.builder.get_object("addDialogApplyButton")
        self.addDialogCancelButton = self.builder.get_object("addDialogCancelButton")
        self.addEntryIP = self.builder.get_object("addEntryIP")
        self.addEntryIP.set_editable(True)
        self.addEntryHostname = self.builder.get_object("addEntryHostname")
        self.addEntryHostname.set_editable(True)
        self.addEntryColor = self.builder.get_object("addEntryColor")
        self.addEntryColor.set_editable(True)


        #### FROM about.glade
        self.aboutDialog = self.builder.get_object("aboutDialog")

        #### FROM help.glade
        self.helpWindow = self.builder.get_object("helpWindow")
        self.helpTextView = self.builder.get_object("helpTextView")

    # CONNECT ALL GUI OBJECTS TO SIGNALS
    def connectAll(self):
        self.window.connect ("destroy", properShutdown)
        self.addB.connect("pressed", self.openAddDialog)
        self.remB.connect("pressed", self.remEntry)
        self.scanB.connect("pressed", self.scan)
        self.dbgB.connect("pressed", self.openDebugger)
        self.sshQJackB.connect("pressed", self.launchSshQJack)
        self.aboutItem.connect("activate", self.openAbout)
        self.helpItem.connect("activate", self.openHelp)
        self.helpWindow.connect("delete-event", self.closeHelp)
        self.addDialogApplyButton.connect("pressed", self.addDialogApply)
        self.addDialogCancelButton.connect("pressed", self.addDialogCancel)
        self.leftDeviceView.connect("row-collapsed", self.linesRedraw)
        self.leftDeviceView.connect("row-expanded", self.linesRedraw)
        self.rightDeviceView.connect("row-collapsed", self.linesRedraw)
        self.rightDeviceView.connect("row-expanded", self.linesRedraw)
        #self.linesArea.connect("draw", self.linesRedraw)


    # POPULATE THE DEVICES VIEW
    def initTreeViews(self):

        # remove these dummy entries once AddButton works and/or scan works
        titer = self.deviceTree.append(None, ["debian-BB1.local 192.168.23.40"])
        self.deviceTree.append(titer, ["input: 2"])
        self.deviceTree.append(titer, ["output: 2"])
        self.deviceTree.append(titer, ["samplerate: 48000"])
        self.deviceTree.append(titer, ["period: 128"])
        self.deviceTree.append(titer, ["tags: guitar"])

        titer = self.deviceTree.append(None, ["debian-BB2.local 192.168.23.41"])
        self.deviceTree.append(titer, ["input: 2"])
        self.deviceTree.append(titer, ["output: 2"])
        self.deviceTree.append(titer, ["samplerate: 48000"])
        self.deviceTree.append(titer, ["period: 128"])
        self.deviceTree.append(titer, ["tags: efx"])

        self.deviceL = Gtk.CellRendererText()
        self.deviceR = Gtk.CellRendererText()
        #self.ipAddr = Gtk.CellRendererText()

        self.columnL = Gtk.TreeViewColumn("device", self.deviceL, text=0) # text=0 means it takes the string from column 0 of the model
        self.columnR = Gtk.TreeViewColumn("device", self.deviceR, text=0) # text=0 means it takes the string from column 0 of the model
        #self.column2 = Gtk.TreeViewColumn("IP", self.ipAddr, text=1)

        self.leftDeviceView.append_column(self.columnL)
        self.rightDeviceView.append_column(self.columnR)

        # init drawing area according to dimensions
        self.curDrawableH = self.leftDeviceView.get_visible_rect().height
        self.linesArea.set_size_request(DRAWING_WIDTH, self.curDrawableH)
        self.linesArea.realize()
        self.drawable = self.linesArea.get_property('window')
        #self.drawable = self.linesArea.window
        # later move these. Todo mettere tutti i metodi di disegno al segnale opportuno di ridimensionamento delle left e right view
        #color = Gdk.Color(red=0, green=0, blue=0)
        #gc = self.linesArea.get_style().white_gc

        #gc = self.linesArea.get_style().new_gc(foreground=None, background=None, font=None, #create graphic context
        #             function=-1, fill=-1, tile=None,
        #             stipple=None, clip_mask=None, subwindow_mode=-1,
        #             ts_x_origin=-1, ts_y_origin=-1, clip_x_origin=-1,
        #             clip_y_origin=-1, graphics_exposures=-1,
        #             line_width=-1, line_style=-1, cap_style=-1,
        #             join_style=-1)



    # per implementare overloading http://stackoverflow.com/questions/6434482/python-function-overloading
    def linesRedraw(self, widget, iter, path, user_param1):
        #redraw, checking if the left and right views changed height
        self.curDrawableH = self.leftDeviceView.get_visible_rect().height
        print "draw"
        self.linesArea.set_size_request(DRAWING_WIDTH, self.curDrawableH)
        cr = widget.get_window().cairo_create()
        cr.set_line_width(2)
        cr.set_source_rgb(0.7, 0.2, 0.0)
        w = DRAWING_WIDTH
        h = self.curDrawableH

        cr.translate(w/2, h/2)
        cr.arc(0, 0, 50, 0, 2*3.14)
        cr.stroke_preserve()

        cr.set_source_rgb(0.3, 0.4, 0.6)
        cr.fill()
#        widget.get_window().draw_line(gc, 0, 0, DRAWING_WIDTH, self.curDrawableH)

    # the same as above, widget is the only argument we need
    def linesRedraw(self, widget, event, user_param1):
        #redraw, checking if the left and right views changed height
        self.curDrawableH = self.leftDeviceView.get_visible_rect().height
        startfrom = self.leftDeviceView.get_visible_rect().width
        print self.curDrawableH
        self.linesArea.set_size_request(DRAWING_WIDTH, self.curDrawableH)
        cr = widget.get_window().cairo_create()
        cr.set_line_width(9)
        cr.set_source_rgb(0.7, 0.2, 0.0)
        w = DRAWING_WIDTH
        h = self.curDrawableH

        w = widget.get_visible_rect().width
        h = widget.get_visible_rect().height

        cr.translate(0, 0)
        cr.line_to(startfrom+1+DRAWING_WIDTH, self.curDrawableH)
        cr.stroke()

#        widget.get_window().draw_line(gc, 0, 0, DRAWING_WIDTH, self.curDrawableH)


    # OPEN ABOUT DIALOG
    def openAbout(self, about):
        res = self.aboutDialog.run()
        self.aboutDialog.hide()

    # OPEN HELP WINDOW
    def openHelp(self, help):
        txtFile = open(HELP_FILE_PATH, 'r')
        try:
            helpTxt = txtFile.read()
        finally:
            txtFile.close()
        self.helpTextView.get_buffer().set_text(helpTxt)
        self.helpWindow.show()

    # OPEN ADD DEVICE DIALOG
    def openAddDialog(self, add):
        res = self.addDialog.run()
        self.aboutDialog.hide()

    # CLOSE HELP WINDOW
    def closeHelp(self, widget, asd):
        self.helpWindow.hide()
        return True

    # ADD DIALOG FUNCTIONS
    def addDialogApply(self, asd):
        # read
        readIP = self.addEntryIP.get_text()
        readHostname = self.addEntryHostname.get_text()
        readColor = self.addEntryColor.get_text()

        # check and append
        if ( isValidHostname(readHostname) ) or ( isValidIP(readIP) ): # if either one not empty
            if not ( isValidHTMLColor(readColor) ):
                readColor = color[0] # default: black
            self.devicesList.append([readHostname, readIP, False, readColor])
            if ( readIP == "" ):
                query = DNSResolveThread(self.devicesList, self.devicesList.get_iter_first(), readHostname, HOSTNAMETYPE)
                threadsList.append(query)
                query.start()
            elif ( readHostname == "" ):
                query = DNSResolveThread(self.devicesList, self.devicesList.get_iter_first(), readIP, IPv4TYPE)
                query.start()
                threadsList.append(query)

        self.addEntryIP.set_text("")
        self.addEntryHostname.set_text("")
        self.addEntryColor.set_text("")
        self.addDialog.hide()


    def addDialogCancel(self, asd):
        self.addDialog.hide()
        return True

    def openDebugger(self, asd):
        dbgProc = subprocess.Popen('wemust-netdbg.py'.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=-1)

    def launchSshQJack(self, asd):
        SshProc = subprocess.Popen('ssh -XY debian@debian-BB1.local qjackctl'.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=-1)
        #stout, sterr = SshProc.communicate()
        # do something?


    # SEND COMMANDS FROM STATUS or DIAGNOSTIC ENTRY TO THE TARGET(s)
    def sendCmd_TCP(self, cmd):
        ipList = []
        # hostnameList = []

        i = self.devicesList.get_iter_first()
        while i:
            if self.devicesList.get_value(i, 2) == True: # if the Address It toggle is ON
                if ( self.devicesList.get_value(i, 1) ):
                    ipList.append( self.devicesList.get_value(i, 1) ) # column 1 = IP
                elif ( self.devicesList.get_value(i, 0) ):
                    ipList.append( self.devicesList.get_value(i, 0) ) # it is ok to append a hostname, socket.connect will handle both IPs and hostnames
                    # hostnameList.append( self.devicesList.get_value(i, 0) )
                else:
                    print("Warning: element " + i + " of the device list has no hostname and IP")


            i = self.devicesList.iter_next(i)

        for r in range(len(ipList)): # starts from 0
            destIP = ipList[r]
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect((destIP, NETDBG_CTRL_PORT))
                sock.send(cmd)
                sock.close()
            except socket.error as sErr:
                print("Warning: socket err " + str(sErr.errno) + ': ' + str(sErr.strerror) )

        # theoretically hostnames are handled as well as IPs it should be safe to delete this
        # for r in range(len(hostnameList)): # starts from 0
        #     destHostname = hostnameList[r]
        #     try:
        #         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        #         sock.connect((destHostname, NETDBG_CTRL_PORT))
        #         sock.send(cmd)
        #         sock.close()
        #     except socket.error as sErr:
        #         print("Warning: socket err " + str(sErr.errno) + ': ' + str(sErr.strerror) )

    # BROADCAST A COMMAND (UNUSED)
    def sendCommand_udp(self, cmd): # discarded
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM,  proto=socket.IPPROTO_UDP) # UDP
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        sock.sendto(cmd, (NETDBG_MCAST_GROUP, NETDBG_CTRL_PORT))

    # SEND A DIAGNOSTIC CHECK (CALLED VIA STATUS BUTTON)
    def defaultDiagnostic(self, button):
        cmd = "ps -e -o pid,uname,pcpu,pmem,comm,args | grep [j]ack"
        self.sendCmd_TCP(cmd)
        cmd = "ps -e -o pid,uname,pcpu,pmem,comm,args | grep [w]emust"
        self.sendCmd_TCP(cmd)



    # ToDo
    def remEntry(self, button):
        model, treeiter = self.devicesView.get_selection().get_selected()
        self.devicesList.remove(treeiter)

    # ToDo
    def scan(self, button):
        pippo = 1




# ------------------------------------------------------------------------ #
# ---------------- NETWORK THREADS HANDLING FUNCTIONS -------------------- #
# ------------------------------------------------------------------------ #

# SPAWN ALL NET THREADS
def threadSpawn():


    # terminal emulator instances
    threadsList.append( NetRxThread(dbgQ, globalQLock, NETDBG_PRINT_PORT) )
    threadsList.append( NetRxThread(jtcliQ, globalQLock, NETDBG_JTCLI_PORT) )
    threadsList.append( NetRxThread(jdQ, globalQLock, NETDBG_JACKD_PORT) )
    threadsList.append( NetRxThread(jtserQ, globalQLock, NETDBG_JTSER_PORT) )

    for t in range(len(threadsList)):
        threadsList[t].start()

# CLOSE ALL THE THREADS
def properShutdown(self):

    quitFlag.set()

    print("Quit...")
    for t in range(len(threadsList)):
        ts = time.time()
        threadsList[t].join(quitTimeout)
        # if the time spent on join() is close to the timeout time it means join spent his time waiting
        # we want to track this down. We can't rely on .isAlive because it always returns true even if join
        # finished fine before the timeout
        if ( time.time() - ts > quitTimeout - 2):
            print("Warning on exit: thread still alive") # maybe it is locked on a resource
            print(threading.enumerate())
    Gtk.main_quit()




# ------------------------------------------------------- #
# ---------------- DNS SEARCH THREAD -------------------- #
# ------------------------------------------------------- #

class DNSResolveThread(threading.Thread):
    def __init__(self, deviceList, dLIter, string, type):
        threading.Thread.__init__(self)
        self.string = string
        self.type = type
        self.deviceList = deviceList
        self.path = self.deviceList.get_path(dLIter)

    def run(self):
        try:
            if ( self.type == HOSTNAMETYPE ):
                try:
                    ip = socket.gethostbyname(self.string)
                    self.deviceList[self.path][1] = ip
                    toKill = self
                except socket.error:
                    toKill = self
                    pass

            elif ( self.type == IPv4TYPE):
                try:
                    hostname = socket.gethostbyaddr(self.string)
                    self.deviceList[self.path][0] = hostname
                    toKill = self
                except socket.error:
                    toKill = self
                    pass
        except (KeyboardInterrupt, SystemExit):
            print("Keyboard Interrupt caught: exiting")
            raise





# ---------------------------------------------------------- #
# ---------------- NETWORK THREAD CLASS -------------------- #
# ---------------------------------------------------------- #

# NetRxThread: thread to receive from the network and hand it to the consoles
class NetRxThread(threading.Thread):
    data = ""
    def __init__(self, termQueue, termQueueLock, port):
        threading.Thread.__init__(self)
        self.termQueue = termQueue
        self.termQueueLock = termQueueLock
        self.RxBufSize = 16384 # a high value can avoid assertion with large messages? gtk_text_view_validate_onscreen: assertion failed: (priv->onscreen_validated)
        self.sock = dbgUDPSockRx(0, NETDBG_MCAST_GROUP, port) # ID still useful?
        self.sock.settimeout(2) # avoid blocking calls to shut down this thread properly
        self.useTimestamp = False

    # ADD TIMESTAMP TO THE INCOMING DATA
    def setUseTimestamp(self, bool):
        self.useTimestamp = bool

    # SEE ABOVE
    def toggleUseTimestamp(self): # Todo: implement toggle in the GUI menus (View)
        self.useTimestamp = not self.useTimestamp

    # SEE ABOVE, GET FUNC
    def getUseTimestamp(self):
        return self.useTimestamp

    # START THREAD
    def run(self):
        while( not quitFlag.is_set() ):
            try:
                self.data, self.addr = self.sock.recvfrom(self.RxBufSize)
                with self.termQueueLock:
                    if self.useTimestamp == True:
                        self.termQueue.append('[' + time.strftime("%H:%M:%S", time.gmtime()) + "] " + self.data + '\n') # alternative witj
                    else:
                        self.termQueue.append(self.addr[0] + ": " + self.data + '\n') # Todo: translate IP into hostname
            except socket.timeout: # timeout has been set for sockets, this guarantees no exc is raised and the thread will go on listening for another timeout cycle
                pass
            except (KeyboardInterrupt, SystemExit):
                print("Keyboard Interrupt caught: exiting")
                raise


# ---------------------------------------------------------------- #
# ---------------- SOCKETS FOR NETWORK THREADS-------------------- #
# ---------------------------------------------------------------- #

class dbgUDPSockRx(socket.socket):
    dbgIP = ""
    dbgPORT = 0
    logFile = ""
    ID = 0 # when initializing it you'd better give a value >0

    def __init__(self, ID, dbgIP, dbgPORT):
        self.ID = ID
        self.dbgIP = dbgIP
        self.dbgPORT = dbgPORT

        socket.socket.__init__(self, family=socket.AF_INET, type=socket.SOCK_DGRAM, proto=socket.IPPROTO_UDP, _sock=None)
        try:
            self.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except AttributeError:
            pass # Some systems don't support SO_REUSEPORT

        try:
            self.bind((dbgIP, dbgPORT))
        except:
            print("Error: could not bind %s:%d" % (dbgIP, dbgPORT))
            self.close()
            raise

        mreq = struct.pack("=4sl", socket.inet_aton(NETDBG_MCAST_GROUP), socket.INADDR_ANY)
        try:
            self.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        except:
            print("Warning: exception: " + str(sys.exc_info()[1]) + ". Connection will not be effective") # Todo: make it a popup dialog
            pass


    def close(self):
        #unsubscribe
        self.setsockopt(socket.SOL_IP, socket.IP_DROP_MEMBERSHIP, socket.inet_aton(self.dbgIP) + socket.inet_aton('0.0.0.0'))
        self.close()


# ------------------------------------------------------------ #
# ------------------------------------------------------------ #
# --------------------- MAIN MAIN MAIN ----------------------- #
# ------------------------------------------------------------ #
# ------------------------------------------------------------ #

ndGUI = connectGUI()
ndGUI.window.show()
tS = threadSpawn()

Gtk.main()
