# wemust defines and classes
# designed for python 2
# currently needs porting for python3 (avahi --> dbus) https://plus.google.com/+JasonDeRose/posts/6uLZn1o3kAj
# Todo:
# - add a class for SABy and related messaging parsing/handling (I want a SAByMsg.getTags, .getNID, .sendADV(name, tags, etc.) and more


import string
import subprocess
import os
import random
import socket
import datetime, time

import dbus, avahi
import threading

## SORT OF MACRO
def uuid(length):
    return ''.join(random.choice(string.ascii_letters+string.digits) for i in range(length))

###### DEFINES ######
WEMUST_GUTS_VERSION = "0.2.0"
USE_NETDBG = 1
VERBOSE_DBG_PRINT = False
USE_LONG_NAMES = 0 # if the devices don't have different names adds a UUID to the hostname
SERVICETYPE = "_wemust._udp"
UUID_LEN = 10
SABy_MCAST_GROUP = "239.255.255.251"
SABy_MCAST_PORT = 1901
SABy_REPLY_TCP_PORT = 8026
SABy_STRING = "SABy-"
SABy_ANNOUNCE_STRING = "ADV"
SABy_SEARCH_STRING = "SRC"
SABy_REPLY_STRING = "REP"
SABy_NID_STRING = "NID:"
SABy_TAG_STRING = "TAG:"
SABy_SR_STRING = "SR:"
SABy_APP_STRING = "APP:"
SABy_MAX_MESSAGE_LEN = 256 # chars

# these are used in .wemustrc config file
DEFAULT_WEMUSTRC_PATH = '/home/debian/wemust/wemust.conf'
WEMUSTRC_TAG_STRING = SABy_TAG_STRING
WEMUSTRC_NID_STRING = SABy_NID_STRING
WEMUSTRC_SR_STRING = SABy_SR_STRING
WEMUSTRC_APP_STRING = SABy_APP_STRING
WEMUSTRC_PEER_STRING = "PEER ADDRESS:"
WEMUST_MAX_NODES = 32 # max nodes on the network (used for socket.listen)
NETDBG_MCAST_GROUP = "239.255.255.252" # SABy +1
NETDBG_BASE_PORT = 1902 # SABy +1
NETDBG_PRINT_PORT = NETDBG_BASE_PORT + 1
NETDBG_JACKD_PORT = NETDBG_BASE_PORT + 2
NETDBG_JTSER_PORT = NETDBG_BASE_PORT + 3
NETDBG_JTCLI_PORT = NETDBG_BASE_PORT + 4
NETDBG_AVPUB_PORT = NETDBG_BASE_PORT + 5 # UNUSED
NETDBG_AVBRO_PORT = NETDBG_BASE_PORT + 6 # UNUSED
NETDBG_CTRL_PORT = NETDBG_BASE_PORT + 7 # Remote control over jackaudio o Beagleboards
if USE_LONG_NAMES == 1:
    SERVICENAME = socket.gethostname() + "-" + uuid(UUID_LEN)
else:
    SERVICENAME = socket.gethostname() # default

MAX_CTRL_CMD_LEN = 200 # 200 allows for long commands (spanning 2-3 terminal lines)

CTRL_CMD_START = 'start'
CTRL_CMD_STOP = 'stop'
CTRL_CMD_HALT = 'halt'
CTRL_CMD_REBOOT = 'reboot'

MAX_PIPE_READLINE_SIZE = 256
HOSTNAMETYPE = "hostname"
IPv4TYPE = "IPv4"
IPv6TYPE = "IPv6"

#################
#### CLASSES ####
#################

### dbgUDPSockTx
class dbgUDPSockTx(socket.socket):   #from http://www.dreamincode.net/forums/topic/209734-using-a-socket-as-stdinstdout/
    dbgIP = ""
    dbgPORT = 0

    def __init__(self, dbgIP, dbgPORT):
        socket.socket.__init__(self, family=socket.AF_INET, type=socket.SOCK_DGRAM, proto=socket.IPPROTO_UDP, _sock=None)
        self.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        self.dbgIP = dbgIP
        self.dbgPORT = dbgPORT

    def write(self, text):
        print(text)
        return self.sendto(text, (self.dbgIP, self.dbgPORT))


# dispatch messages from processes to the related debug sockets
def dispatcherThread(pipe,socket):
    while 1:
        buf = pipe.readline() # read
        while buf:
            socket.write(buf.strip()) # send (remove trailing \n with strip)
            buf = pipe.readline()
        time.sleep(1)

### SAByParams
class SAByParams():
    app = ""
    nid = ""
    tag = ""
    sr = ""

    def setFromFile(self, filename):
        toParse = []
        for line in open(filename):
            li = line.strip()
            if not li.startswith("#"):
                toParse.append(line.rstrip())
        for line in toParse:
            if line.startswith(WEMUSTRC_TAG_STRING):
                self.tag = str( line[len(WEMUSTRC_TAG_STRING):len(line)] )
            if line.startswith(WEMUSTRC_NID_STRING):
                self.nid = str( line[len(WEMUSTRC_NID_STRING):len(line)] )
            if line.startswith(WEMUSTRC_SR_STRING):
                self.sr = str( line[len(WEMUSTRC_SR_STRING):len(line)] )
        self.app = __file__

    def setFromAnnounceMsg(self, msg):
        toParse = msg.split()
        for line in toParse:
            if line.startswith(SABy_TAG_STRING):
                self.tag = str( line[len(SABy_TAG_STRING):len(line)] )
            if line.startswith(SABy_NID_STRING):
                self.nid = str( line[len(SABy_NID_STRING):len(line)] )
            if line.startswith(SABy_SR_STRING):
                self.sr = str( line[len(SABy_SR_STRING):len(line)] )
            if line.startswith(SABy_APP_STRING):
                self.app = str( line[len(SABy_APP_STRING):len(line)] )

    def setParams(self, app, nid, tag, sr):
        self.app = str(app)
        self.nid = str(nid)
        self.tag = str(tag)
        self.sr = str(sr)

    def getAllParams(self):
        return self.app, self.nid, self.tag, self.getSampRateOnly(), self.getPeriodOnly()

    # Always be aware that sr is a string of the kind '44100|128', thus to get the samplerate or period only there are two different methods getSampRateOnly e getPeriodOnly
    def getSr(self):
        return self.sr

    def getSampRateOnly(self):
        return self.sr.split('|')[0]

    def getPeriodOnly(self):
        return self.sr.split('|')[1]

    def getSearchMessage(self): # tags are optional
        return SABy_STRING + SABy_SEARCH_STRING + "\r\n" + "APP:\r\n" + "NID:\r\n" + "TAG:" + self.tag + "\r\n" + "SMP:|\r\n"

    def getAnnounceMessage(self): # tags are optional
        return SABy_STRING + SABy_ANNOUNCE_STRING + "\r\n" + "APP:" + self.app + "\r\n" + "NID:" + self.nid + "\r\n" + "TAG:" + self.tag + "\r\n" + "SMP:" + self.sr + "\r\n"

    def getReplyMessage(self): # tags are optional
        return SABy_STRING + SABy_REPLY_STRING + "\r\n" + "APP:" + self.app + "\r\n" + "NID:" + self.nid + "\r\n" + "TAG:" + self.tag + "\r\n" + "SMP:" + self.sr + "\r\n"

    def search(self):
        UDPSend(self.getSearchMessage(), SABy_MCAST_GROUP, SABy_MCAST_PORT)

    def announce(self):
        UDPSend(self.getAnnounceMessage(), SABy_MCAST_GROUP, SABy_MCAST_PORT)

    def reply(self, addr_tuple):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((addr_tuple[0], SABy_REPLY_TCP_PORT)) # addr[0] = address string, addr[1] = originating port
        s.send( self.getReplyMessage() )
        s.close()




##### FUNCTIONS #####

def getIPfromWEMUSTRC():
    toParse = []
    for line in open(DEFAULT_WEMUSTRC_PATH):
        li = line.strip()
        if not li.startswith("#"):
            toParse.append(line.rstrip())
    for line in toParse:
        if line.startswith(WEMUSTRC_PEER_STRING):
             return str( line[len(WEMUSTRC_PEER_STRING):len(line)] )
    return ""

def UDPSend(cmd, IP, port): # discarded
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM,  proto=socket.IPPROTO_UDP) # UDP
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
    sock.sendto(cmd, (IP, port))

def which(binname):
    p = subprocess.Popen(['which', binname], stdout=subprocess.PIPE)
    out, err = p.communicate()
    return out

def timestamp():
    #return datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
    #return time.time()
    return datetime.datetime.fromtimestamp(time.time()).strftime('%H:%M:%S')

# debug announce thread - NOT USED AT THE MOMENT
def dbg_announce(basePort, devid, delay):
    announce_msg = "ID: %s, DBG-BASE-PORT: %d" % (devid, basePort)
    dbg_announce_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP) # UDP
    dbg_announce_sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2) # MULTICAST, alcuni dubbi https://wiki.python.org/moin/UdpCommunication, vd anche http://pymotw.com/2/socket/multicast.html
    dbg_announce_sock.sendto(announce_msg, (NETDBG_MCAST_GROUP,NETDBG_BASE_PORT))

# required once before starting using wemust.dbg_print
def wInit():
    global netdbg_print_sock
    netdbg_print_sock = dbgUDPSockTx(NETDBG_MCAST_GROUP,NETDBG_PRINT_PORT)
    try:
        os.remove("wemust-start-print-log.txt") # if log file exists remove
    except OSError:
        pass
    return netdbg_print_sock

# network print function for debug
def dbg_print(txt):
    if USE_NETDBG == 1:
        if VERBOSE_DBG_PRINT == True:
            txt = '[ID:%s][t:%s] ' % (SERVICENAME, timestamp()) + txt
        netdbg_print_sock.sendto(txt, (NETDBG_MCAST_GROUP, NETDBG_PRINT_PORT))
    else:
        with open("wemust-start-print-log.txt", "a") as myfile:
            myfile.write(txt)


# portCheck: this version relies on shell commands, not very portable...
def portCheck(port):
    chk = subprocess.check_output("lsof -nP -S 5 -Fpc -i:" + str(port) + " | tr -d \'\n\'",shell=True) #tr is needed to remove newlines (multiple lines test give bash error)
    if chk == '': # another method could be looking for alphanums in case any newline or special character would remain
        return True
    else:
        return False

# avahi resolved service print
def resolved_serv_print(*args):
    print("service resolved")
    print("name:", args[2])
    print("address:", args[7])
    print("port:", args[8])

# launch jt client (finally!) if avahi found a wemust service
def jt_cli_spawn(*args):
    PEER_IP = args[7]
    PEER_PORT = args[8] # NON SERVE???
    dbg_print("Launching " + JTBIN + " -c " + PEER_IP + " -o " + jt_cli_port_offs + " --clientname " + CLIENTNAME)
    jtbin_client_proc = subprocess.Popen(JTBIN + " -c " + PEER_IP + " -o " + jt_cli_port_offs + " --clientname " + CLIENTNAME, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=-1)
    jt_cli_dt = threading.Thread(target=wemust.dispatcherThread, args=(jtbin_client_proc.stdout, jt_cli_dbgs))
    jt_cli_dt.daemon = True # thread dies with the program
    jt_cli_dt.start()
    jt_cli_port_offs += 1 # ToDo: check incremento +1 o +2?
    jtbin_client_proc.wait()
    dbg_print("Apparently jacktrip Exited or Died, with exit code %d" % jtbin_client_proc.returncode)

# print avahi errors
def print_error(*args):
    print("Avahi-wemust error_handler")
    print(args[0])

# handler for avahi when a wemust service is found
def found_handler(interface, protocol, name, stype, domain, flags):
    dbg_print("Found service " + name + " type " + stype + " domain " + domain)
    # CHKPREVCONNECT Todo: check if already connected to this service (needed: a global list of IPs)
    if flags & avahi.LOOKUP_RESULT_LOCAL: # local service, skip
            pass
    else:
        server.ResolveService(interface, protocol, name, stype, domain, avahi.PROTO_UNSPEC, dbus.UInt32(0), reply_handler=jt_cli_spawn, error_handler=print_error) # launch JACKTRIP!
