#include <stdio.h>
#include <unistd.h>
#include "dm3730_temp_sensor.c"

#define DEFAULT_READ_INTERVAL_SECS 1
#define ONE_SEC 1
#define ONE_MIN 60
#define CONVERSIONTIMEUSECS 100000 // 100ms works fine

void dm3730_get_temp_start();
int dm3730_get_temp_check(float *temp_val);
float dm3730_get_temp_finish();

void help_print() {
  fprintf(stdout, "\nget_temp for DM3730 ARM processors\n");
  fprintf(stdout, "usage: get_temp [-m|-s] \nwhere -m = read interval 1 minute, -s = read interval 1 second\n");
}

int main(int argc, char *argv[]) {
  int read_interval = 1;
  float this_temp;
  int waiting_on_conversion = 0;
  int opt = 0;

  /* INPUT OPTION PARSING */
  if ( argc == 1 ) {
    fprintf(stdout, "\nDefault: print every %d sec\n", DEFAULT_READ_INTERVAL_SECS);
    read_interval = DEFAULT_READ_INTERVAL_SECS;
  }
  else if ( argc > 2 ) {
    fprintf(stdout, "\nInvalid option received\n");
    help_print();
    return -1;
  }
  else {
    while ((opt = getopt(argc, argv, "ms")) != -1) {
      switch(opt) {
      case 'm':
        read_interval = ONE_MIN;
        break;
      case 's':
        read_interval = ONE_SEC;
        break;
      case 'h':
        help_print();
        return -1;
      case '?':
        /* Case when user enters the command as
         * $ ./cmd_exe -i
         */
        if (optopt == 'm')
          read_interval = ONE_MIN;
        else if (optopt == 's')
          read_interval = ONE_SEC;
        else if (optopt == 'h') {
          help_print();
          return -1;
        }
        else {
          fprintf(stdout, "\nInvalid option received\n");
          help_print();
          return -1;
        }
        break;
      }
    }
  }
  /* LOOP */
  while(1) {
    if(!waiting_on_conversion) {
      dm3730_get_temp_start();
      waiting_on_conversion = 1;
    }
    else {
      if(dm3730_get_temp_check(&this_temp)) { //Still waiting
      }
      else { //Conversion done
        fprintf(stdout, "T = %.1f deg\r\n", this_temp);
        fflush(stdout);
        waiting_on_conversion = 0;
	sleep(read_interval);
      }
    }
    usleep(CONVERSIONTIMEUSECS);
  }
  return 0;
}

